@@ -9,8 +9,8 @@ use crate::{
     units::{
         ControlHash, FullUnit, PreUnit, SignedUnit, UncheckedSignedUnit, Unit, UnitCoord, UnitStore,
     },
-    Config, Data, DataIO, Hasher, Index, MultiKeychain, NodeCount, NodeIndex, OrderedBatch,
-    Receiver, Round, Sender, SessionId, Signature, Signed, SpawnHandle, UncheckedSigned,
+    Config, Data, DataProvider, FinalizationHandler, Hasher, Index, MultiKeychain, NodeCount,
+    NodeIndex, Receiver, Round, Sender, SessionId, Signature, Signed, SpawnHandle, UncheckedSigned,
 };
 use futures::{
     channel::{mpsc, oneshot},
@@ -102,12 +102,13 @@ impl<H: Hasher, D: Data, S: Signature> TryFrom<UnitMessage<H, D, S>>
     }
 }
 
-struct Runway<'a, H, D, MK, DP>
+struct Runway<'a, H, D, MK, DP, FH>
 where
     H: Hasher,
     D: Data,
     MK: MultiKeychain,
-    DP: DataIO<D>,
+    DP: DataProvider<D>,
+    FH: FinalizationHandler<D>,
 {
     missing_coords: HashSet<UnitCoord>,
     missing_parents: HashSet<H::Hash>,
@@ -125,7 +126,8 @@ where
     tx_consensus: Sender<NotificationIn<H>>,
     rx_consensus: Receiver<NotificationOut<H>>,
     ordered_batch_rx: Receiver<Vec<H::Hash>>,
-    data_io: DP,
+    data_provider: DP,
+    finalization_handler: FH,
     after_catch_up_delay: bool,
     starting_round_sender: Option<oneshot::Sender<Round>>,
     starting_round_value: Round,
@@ -134,13 +136,21 @@ where
     exiting: bool,
 }
 
-struct RunwayConfig<'a, H: Hasher, D: Data, DP: DataIO<D>, MK: MultiKeychain> {
+struct RunwayConfig<
+    'a,
+    H: Hasher,
+    D: Data,
+    DP: DataProvider<D>,
+    FH: FinalizationHandler<D>,
+    MK: MultiKeychain,
+> {
     node_ix: NodeIndex,
     session_id: SessionId,
     n_members: NodeCount,
     max_round: Round,
     keychain: &'a MK,
-    data_io: DP,
+    data_provider: DP,
+    finalization_handler: FH,
     alerts_for_alerter: Sender<Alert<H, D, MK::Signature>>,
     notifications_from_alerter: Receiver<ForkingNotification<H, D, MK::Signature>>,
     tx_consensus: Sender<NotificationIn<H>>,
@@ -153,14 +163,15 @@ struct RunwayConfig<'a, H: Hasher, D: Data, DP: DataIO<D>, MK: MultiKeychain> {
     salt: u64,
 }
 
-impl<'a, H, D, MK, DP> Runway<'a, H, D, MK, DP>
+impl<'a, H, D, MK, DP, FH> Runway<'a, H, D, MK, DP, FH>
 where
     H: Hasher,
     D: Data,
     MK: MultiKeychain,
-    DP: DataIO<D>,
+    DP: DataProvider<D>,
+    FH: FinalizationHandler<D>,
 {
-    fn new(config: RunwayConfig<'a, H, D, DP, MK>) -> Self {
+    fn new(config: RunwayConfig<'a, H, D, DP, FH, MK>) -> Self {
         let n_members = config.n_members;
         let threshold = (n_members * 2) / 3 + NodeCount(1);
         let max_round = config.max_round;
@@ -180,7 +191,8 @@ where
             tx_consensus: config.tx_consensus,
             rx_consensus: config.rx_consensus,
             ordered_batch_rx: config.ordered_batch_rx,
-            data_io: config.data_io,
+            data_provider: config.data_provider,
+            finalization_handler: config.finalization_handler,
             after_catch_up_delay: false,
             starting_round_sender: Some(config.starting_round_sender),
             starting_round_value: 0,
@@ -580,7 +592,7 @@ where
 
     async fn on_create(&mut self, u: PreUnit<H>) {
         debug!(target: "AlephBFT-runway", "{:?} On create notification.", self.index());
-        let data = self.data_io.get_data();
+        let data = self.data_provider.get_data().await;
         let full_unit = FullUnit::new(u, data, self.session_id);
         let signed_unit = Signed::sign(full_unit, self.keybox).await;
         self.store.add_unit(signed_unit.clone(), false);
@@ -670,8 +682,8 @@ where
         }
     }
 
-    fn on_ordered_batch(&mut self, batch: Vec<H::Hash>) {
-        let batch = batch
+    async fn on_ordered_batch(&mut self, batch: Vec<H::Hash>) {
+        let data_iter: Vec<_> = batch
             .iter()
             .map(|h| {
                 self.store
@@ -681,9 +693,10 @@ where
                     .data()
                     .clone()
             })
-            .collect::<OrderedBatch<D>>();
-        if let Err(e) = self.data_io.send_ordered_batch(batch) {
-            error!(target: "AlephBFT-runway", "{:?} Error when sending batch {:?}.", self.index(), e);
+            .collect();
+
+        for d in data_iter {
+            self.finalization_handler.data_finalized(d).await;
         }
     }
 
@@ -769,7 +782,7 @@ where
                 },
 
                 batch = self.ordered_batch_rx.next() => match batch {
-                    Some(batch) => self.on_ordered_batch(batch),
+                    Some(batch) => self.on_ordered_batch(batch).await,
                     None => {
                         error!(target: "AlephBFT-runway", "{:?} Ordered batch stream closed.", index);
                         break;
@@ -812,10 +825,11 @@ pub(crate) struct RunwayIO<H: Hasher, D: Data, MK: MultiKeychain> {
     pub(crate) resolved_requests: Sender<Request<H>>,
 }
 
-pub(crate) async fn run<H, D, MK, DP, SH>(
+pub(crate) async fn run<H, D, MK, DP, FH, SH>(
     config: Config,
     keychain: MK,
-    data_io: DP,
+    data_provider: DP,
+    finalization_handler: FH,
     spawn_handle: SH,
     runway_io: RunwayIO<H, D, MK>,
     mut exit: oneshot::Receiver<()>,
@@ -823,7 +837,8 @@ pub(crate) async fn run<H, D, MK, DP, SH>(
     H: Hasher,
     D: Data,
     MK: MultiKeychain,
-    DP: DataIO<D>,
+    DP: DataProvider<D>,
+    FH: FinalizationHandler<D>,
     SH: SpawnHandle,
 {
     let (tx_consensus, consensus_stream) = mpsc::unbounded();
@@ -883,7 +898,8 @@ pub(crate) async fn run<H, D, MK, DP, SH>(
 
     let runway_config = RunwayConfig {
         keychain: &keychain,
-        data_io,
+        data_provider,
+        finalization_handler,
         alerts_for_alerter,
         notifications_from_alerter,
         tx_consensus,
