@@ -1,14 +1,11 @@
-#[cfg(test)]
+#![cfg(test)]
+mod alerts;
 mod byzantine;
-#[cfg(test)]
 mod consensus;
-#[cfg(test)]
 mod crash;
-#[cfg(test)]
+mod creation;
+mod dag;
 pub(crate) mod mock;
-#[cfg(test)]
 mod rmc;
-#[cfg(test)]
 pub(crate) mod signed;
-#[cfg(test)]
 mod unreliable;
